const { Group, Field, randomId, parseArgs } = require('@igomoon/hubspot-fields-js');
const path = require('path');
const { logger } = require('@hubspot/cli-lib/logger')
const { deepExtend } = require('../js-toolkit/util/obj');

/**
 * Load File uncached
 */
const requireUncached = (path) => { 
	delete require.cache[require.resolve(path)];
	return require(path);
}

/**
 * Load Field and Clear Cache from fields folder in src
 */
let globalFieldsDirPath = null
const setGlobalFieldsDirPath = (path) => {
	globalFieldsDirPath = path
	return globalFieldsDirPath
}

let usePackageFields = true
const setEnablePackageFields = (val) => {
	usePackageFields = !!val
	return usePackageFields
}

const loadFromFieldsDir = (module, lookInPackage = true) => {

	if (module.startsWith('#/') && lookInPackage) {
		return loadFromPackageFieldsDir(module.substr(2))
	}

	if (!globalFieldsDirPath) {
		if (usePackageFields && lookInPackage) {
			return loadFromPackageFieldsDir(module)
		}
		throw new Error('Global Fields Directory Path not set. Use setGlobalFieldsDirPath(path).')
	}

	try {
		let pathToField = path.resolve(globalFieldsDirPath, module);
		return requireUncached(pathToField);
	} catch (e) {
		if (usePackageFields && lookInPackage) {
			return loadFromPackageFieldsDir(module)
		}
		logger.error(`Could not find file "${module}" in assigned fields dir: ${globalFieldsDirPath}`, e.requireStack || '')
		throw e
    }
}

const loadFromPackageFieldsDir = (module) => {
	if (!usePackageFields) {
		throw new Error('Use fields from the package directory is disabled. Use setEnablePackageFields(true) to enable.')
	}
	try {
		let pathToField = path.resolve(__dirname + '/lib', module);
		return requireUncached(pathToField);
	} catch (e) {
		logger.error("Could not find file in package fields dir: " + module, e.requireStack)
		throw e
    }
}

/**
 * Library
 */
let fieldsLibrarySrc = null
const setFieldLibrarySrc = (path) => {
	fieldsLibrarySrc = path
	return fieldsLibrarySrc
}

let usePackageLibrary = true
const setEnablePackageLibrary = (val) => {
	usePackageLibrary = !!val
	return usePackageLibrary
}

const getLibrary = () => {
	// Get Package Config
	let packageLibrary = usePackageLibrary ? require('./library') : {}
	if (!fieldsLibrarySrc) { return packageLibrary }
	// Merge in Custom Library
	let customLibrary = requireUncached(fieldsLibrarySrc)
	return deepExtend(packageLibrary, customLibrary)
}

const loadFromLibrary = (selector) => {
	if (!fieldsLibrarySrc) { 
		throw new Error('Fields Library Source Path not set. Use setFieldLibrarySrc(path)')
	}
	let library = getLibrary()
	if (typeof selector === 'string' || selector instanceof String) {
		let selectorPath = selector.split('.');
		let first = [...selectorPath].shift().toUpperCase();
		if (library[first]) {
			return selectorPath.reduce((o, k) => o[k.toUpperCase()], library);
		}
	}
	return null;
}

const LibraryField = class {
	constructor(path) {
		let fetchedField = loadFromLibrary(path)
		if (!fetchedField) {
			throw new Error(`Library Field "${path}" not found. Make sure it is defined in ${fieldsLibrarySrc}`)
		}
		if (typeof fetchedField !== 'function') {
			throw new Error(`Library Field "${path}" must return a function or class constructor.`)
		}
		const isClass = fn => /^\s*class/.test(fn.toString());
		return (isClass(fetchedField) ? new fetchedField() : fetchedField())
	}
}

/**
 * Config
 */
let fieldsConfigSrc = null
const setFieldConfigSrc = (path) => {
	fieldsConfigSrc = path
	return fieldsConfigSrc
}

let usePackageConfig = true
const setEnablePackageConfig = (val) => {
	usePackageConfig = !!val
	return usePackageConfig
}
 
const getConfig = () => {
	// Get Package Config
	let packageConfig = usePackageConfig ? require('./config') : {}
	if (!fieldsConfigSrc) { return packageConfig }
	// Merge in Custom Config
	let customConfig = requireUncached(fieldsConfigSrc)
	return deepExtend(packageConfig,customConfig)
}

const getFromConfig = (key, defaultValue) => {
	if (!fieldsConfigSrc) { 
		throw new Error('Fields Config Source Path not set. Use setFieldConfigSrc(path)')
	}
	let configVars = getConfig()
	try {
        if(configVars[key] != undefined) { return configVars[key]; }
		if (defaultValue != 'undefined') { return defaultValue }
        throw new Error('No configuration option found for key: ' + key);
    } catch(e) {
        logger.error('Could not read configuration item:\n' + e);
		throw e
    }
}

/**
 * ID Helper
 */
let fieldJsGeneratedIds = {}
const getOrCreateId = (key) => {
	if (!fieldJsGeneratedIds[key]) { fieldJsGeneratedIds[key] = randomId()}
	return fieldJsGeneratedIds[key];
}

/**
 * Exports
 */
module.exports = {
	requireUncached,
	// Fields Dir
	setGlobalFieldsDirPath,
	setEnablePackageFields,
	loadFromFieldsDir,
	loadFromPackageFieldsDir,
	// Library
	setFieldLibrarySrc,
	setEnablePackageLibrary,
	library : getLibrary,
	loadFromLibrary,
	LibraryField,
	// Config
	setFieldConfigSrc,
	setEnablePackageConfig,
	getConfig,
	config: getFromConfig,
	// Id Helper
	getOrCreateId,
	// Field JS Package
	Group,
	Field,
	randomId,
	parseArgs
};